#include "../include/mpi.h"

int           MPI_errno;

#define MAX_PENDING 100
static int           curpending = 0;
static _MPI_pendings PendingMsgs[MAX_PENDING];

/*D
   MPI_main - Name of main program for MPI applications

   Note:
   Use "MPI_main" instead of "main" in C programs, and instead of
   "PROGRAM <name>" in Fortran programs.

   Synopsis:
   (C)
   int MPI_main( argc, argv )
   int argc;
   char **argv;

   Synopsis:
   (Fortran)
   integer function MPI_main()
D*/
int main( argc, argv )
int  argc;
char *argv;
{
    extern int MPI_main();
    return PICall( MPI_main, argc, argv );
}

/*@
  MPI_stopall - stop all processes

  Input Parameter:
. rc - return code to pass back to calling environment
@*/
void MPI_stopall( rc )
{
    SYexitall( (char *)0, rc );
}

/* 
   The pending structure is used to keep track of nonblocking send/receive
   parameters between the initiation and the MPI_wait.
 */
/*
  Internal routine to get pending structure
 */
_MPI_pendings *_MPI_GetP( kind, flavor )
_MPI_Kind   kind;
_MPI_Flavor flavor;
{
    _MPI_pendings *p = 0;
    if (curpending < MAX_PENDING) {
	p         = PendingMsgs + curpending++;
	p->kind   = kind;
	p->flavor = flavor;
	}
    return p;
}

/* 
   Internal routine to free pending structure 
 */
void _MPI_FreeP( p )
_MPI_pendings *p;
{
    p->kind = _MPI_EMPTY;
    if (PendingMsgs + curpending - 1 == p) curpending--;
}

/* Still need to insure that only maxlis members are set */
/*@
  MPI_infog - Determine the number of processes in a group and return
  the PID numbers of the group members

  Input Parameters:
.  gid    - group id
.  maxlis - size of plist
.  plist  - integer array to hold the members of the group

  Returns:
  The number of members in the group, or -1 on error.
@*/ 
int MPI_infog(gid,maxlis,plist) 
int gid, maxlis, *plist;
{
    if (maxlis >= PSNUMNODES((ProcSet *)gid)) {
	PSPROCLIST((ProcSet *)gid,plist);
	}
    return PSNUMNODES((ProcSet *)gid);
}


/*@
  MPI_wait - Block until a nonblocking send or receive operation has completed

  Input Parameter:
. msg_id - id returned by a nonblocking send or receive routine (of any type)

  Returns:
  Number of bytes sent or received, or -1 on error.
@*/
int MPI_wait( msg_id )
int msg_id;
{
    _MPI_pendings *p = (_MPI_pendings *)msg_id;
    int rc;
    
    if (p->flavor == _MPI_CONTIG) {
	if (p->kind == _MPI_RECV) {
	    RECVWAITNOMEM( p->type, p->buf, p->len, MSG_OTHER, p->rid );
	    rc = RECVLEN();
	    }
	else if (p->kind == _MPI_SEND) {
	    SENDWAITNOMEM( p->type, p->buf, p->len, p->partner, 
			   MSG_OTHER, p->rid );
	    rc = p->len;
	    }
	else {
	    MPI_errno = 3;
	    return -1;
	    }
	}
    else if (p->flavor == _MPI_STRIDE) {
	if (p->kind == _MPI_RECV) {
	    RECVVECWAIT( p->type, p->buf, p->lenblk, p->stride, p->nblks, 
			 MSG_OTHER, p->rid );
	    rc = RECVLEN();
	    }
	else if (p->kind == _MPI_SEND) {
	    SENDVECWAIT( p->type, p->buf, p->lenblk, p->stride, p->nblks, 
			 p->partner, MSG_OTHER, p->rid );
	    rc = p->len;
	    }
	else {
	    MPI_errno = 3;
	    return -1;
	    }
	}
    else if (p->flavor == _MPI_GENERAL) {
	if (p->kind == _MPI_RECV) {
	    RECVVWAIT( p->type, &p->bv, p->rid );
	    rc = RECVLEN();
	    }
	else if (p->kind == _MPI_SEND) {
	    SENDVWAIT( p->type, &p->bv, p->partner, p->rid );
	    rc = p->len;
	    }
	else {
	    MPI_errno = 3;
	    return -1;
	    }
	}
    _MPI_FreeP( p );
    return rc;
}

/*@
   MPI_probe - Check for pending messages

   Input Parameters:
.  source - the PID of the process sending the message
.  type   - the message type   

   Returns:
   Length of pending message if available, else -1.
@*/
int MPI_probe( source, type )
int source, type;
{
    if (ASYNCPROBE(type))
	return RECVLEN();
    return -1;
}

/*@
  MPI_stats - Check the status of a nonblocking send or receive

  Input Parameter:
. msg_id - Message id returned by a call to a nonblocking send or receive  

  Returns:
  Length of available message if it is pending, else -1.
@*/
int MPI_stats( msg_id ) 
int msg_id;
{
    _MPI_pendings *p = (_MPI_pendings *)msg_id;
    if (p->kind == _MPI_RECV) {
	if (ASYNCDONE(p->rid))
	    return RECVLEN();
	}
    else {
	if (ASYNCDONE(p->sid)) {
	    return p->len;
	    }
	}
    return -1;
}

/*@
  MPI_cancel - Cancel a previously initiated nonblocking send or receive

  Input Parameter:
. msgid - Message id returned by a call to a nonblocking send or receive  
@*/
int MPI_cancel( msgid )
int msgid;
{
return 0;
}

/*@
  MPI_error - Determine the error status

  Returns:
  Integer giving the error status for the preceding call to an MPI routine.
@*/
int MPI_error()
{
return MPI_errno;
}

/*@
  MPI_etext - Return the text corresponding to an error value

  Input Parameter:
. ierrno - Value returned by MPI_error
@*/
char *MPI_etext( ierrno )
int ierrno;
{
switch (ierrno) {
    case 0: return "No error";
    case 1: return "Invalid PID used";
    case 2: return "Invalid GID used";
    case 3: return "Invalid MSGID used";
    case 4: return "Invalid CCID used";
    case 5: return "Invalid GCPID used";
    case 6: return "Invalid message buffer size";
    case 7: return "Invalid stride in MPI_spack or MPI_sunpack";
    case 8: return "Invalid blocksize in MPI_gpack or MPI_gunpack";
    case 9: return "Invalid data item size in MPI_gpack or MPI_gunpack";
    case 10:return "System buffer overflow";
    case 11:return "Too many communication contexts";
    case 12:return "Too many group contexts";
    case 13:return "Can not delete root group";
    case 14:return "Unknown mode";
    default:return "Invalid error";
    }
}

/*@
  MPI_infoc - Get information about valid communication contexts

  Input parameter:
. maxlis - maximum size of ilist

  Output parameter:
. ilist  - array containing communication context ID numbers
@*/  
int MPI_infoc( maxlis, ilist )
int maxlis, *ilist;
{
MPI_errno = 0;
if (maxlis > 0)
    ilist[0] = 0;
return 1;
}

/*@
  MPI_newc - Create a new communication context

  Returns:
  The id number of a new communication context, or -1 on error.
@*/
int MPI_newc()
{
MPI_errno = 11;
return -1;
}

/*@
  MPI_popc - Reestablish former communication context

  Returns:
  The id number of a new communication context, or -1 on error.
@*/
int MPI_popc()
{  
MPI_errno = 11;
return -1;
}

/*@
  MPI_pushc - Establish a new communication context

  Input Parameter:
. ccid - the ID number of communication context to establish

  Returns:
  0 on success or -1 on error.
@*/
int MPI_pushc( ccid )
int ccid;
{  
MPI_errno = 11;
return -1;
}

/*@
   MPI_machine - Get machine name, version, and related information

   Returns:
   A character string giving the name, etc., of the machine.
@*/
char *MPI_machine( )
{
static char buf[256];

SYGetHostName( buf, 255 );
return buf;
}

/*@
  MPI_infonm - Get information on the machine configuration

  Input Parameter:
. maxlis - the size of ilist

  Output Parameter:
. ilist - integer array containing information about the system.
$          ilist[0] - number of physical processors in the machine
$          ilist[1] - total number of processors in the machine
@*/
int MPI_infonm( maxlis, ilist )
int maxlis, *ilist;
{
int cnt=0;

MPI_errno = 0;
if (maxlis > 0) 
    ilist[cnt++] = NUMNODES;
if (maxlis > 1)
    ilist[cnt++] = NUMNODES;
return cnt;
}
