#include "../include/mpi.h"

extern _MPI_pendings *_MPI_GetP( );
extern void _MPI_FreeP( );

/*@
    MPI_cbsend - Blocking contiguous send

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   len  - length in bytes of message

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_cbsend( buf, dest, type, len )
void *buf;
int  dest, type, len;
{
SENDSYNCNOMEM(type,buf,len,dest,MSG_OTHER);
return len;
}

/*@
    MPI_cnsend - Nonblocking contiguous send

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   len  - length in bytes of message

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_cnsend( buf, dest, type, len )
void *buf;
int  dest, type, len;
{
_MPI_pendings *p;

p  = _MPI_GetP(_MPI_SEND,_MPI_CONTIG);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_SEND
p->buf     = buf;
p->type    = type;
p->len     = len;
p->partner = dest;
#endif
SENDASYNCNOMEM(type,buf,len,dest,MSG_OTHER,p->sid);
return (int)p;
}

/*@
    MPI_cbrecv - Blocking contiguous receive

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message

    Returns:
    Number of bytes actually received.
@*/
int MPI_cbrecv( buf, source, type, maxlen )
void *buf;
int  source, type, maxlen;
{
RECVSYNCNOMEM(type,buf,maxlen,MSG_OTHER);
return RECVLEN();
}

/*@
    MPI_cnrecv - Nonblocking contiguous receive

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_cnrecv( buf, source, type, maxlen )
void *buf;
int  source, type, maxlen;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_RECV,_MPI_CONTIG);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_RECV
p->buf     = buf;
p->type    = type;
p->len     = maxlen;
p->partner = source;
#endif
RECVASYNCNOMEM(type,buf,maxlen,MSG_OTHER,p->rid);
return (int)p;
}

/*@
    MPI_cbsendrr - Blocking contiguous send for ready receivers

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   len  - length in bytes of message

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_cbsendrr( buf, dest, type, len )
void *buf;
int  dest, type, len;
{
SENDSYNCNOMEMFORCE(type,buf,len,dest,MSG_OTHER);
return len;
}

/*@
    MPI_cnsendrr - Nonblocking contiguous send for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_cnsendrr( buf, dest, type, len )
void *buf;
int  dest, type, len;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_SEND,_MPI_CONTIG);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_SEND
p->buf     = buf;
p->type    = type;
p->len     = len;
p->partner = dest;
#endif
SENDASYNCNOMEMFORCE(type,buf,len,dest,MSG_OTHER,p->sid);
return (int)p;
}

/*@
    MPI_cbrecvrr - Blocking contiguous receive for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message

    Returns:
    Number of bytes actually received.
@*/
int MPI_cbrecvrr( buf, source, type, maxlen )
void *buf;
int  source, type, maxlen;
{
RECVSYNCNOMEMFORCE(type,buf,maxlen,MSG_OTHER);
return RECVLEN();
}

/*@
    MPI_cnrecvrr - Nonblocking contiguous receive for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_cnrecvrr( buf, source, type, maxlen )
void *buf;
int  source, type, maxlen;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_RECV,_MPI_CONTIG);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_RECV
p->buf     = buf;
p->type    = type;
p->len     = maxlen;
p->partner = source;
#endif
RECVASYNCNOMEMFORCE(type,buf,maxlen,MSG_OTHER,p->rid);
return (int)p;
}

/*@
    MPI_csend - Draft standard contiguous send

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to send
.   dest - destination processor
.   type - message type
.   len  - length in bytes of message

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_csend( mode, buf, dest, type, len )
char *mode;
void *buf;
int  dest, type, len;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_cbsend(buf,dest,type,len);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_cnsend(buf,dest,type,len);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_cbsend(buf,dest,type,len);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_crecv - Draft standard contiguous receive

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_crecv( mode, buf, source, type, maxlen )
char *mode;
void *buf;
int  source, type, maxlen;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_cbrecv(buf,source,type,maxlen);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_cnrecv(buf,source,type,maxlen);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_cbrecv(buf,source,type,maxlen);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_csendrr - Friendly contiguous send for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to send
.   dest - destination processor
.   type - message type
.   len  - length in bytes of message

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_csendrr( mode, buf, dest, type, len )
char *mode;
void *buf;
int  dest, type, len;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_cbsendrr(buf,dest,type,len);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_cnsendrr(buf,dest,type,len);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_cbsendrr(buf,dest,type,len);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_crecvrr - Friendly contiguous receive for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_crecvrr( mode, buf, source, type, maxlen )
char *mode;
void *buf;
int  source, type, maxlen;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_cbrecvrr(buf,source,type,maxlen);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_cnrecvrr(buf,source,type,maxlen);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_cbrecvrr(buf,source,type,maxlen);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_cbsendh - Blocking contiguous heterogeneous send

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   len  - length in bytes of message
.   datatype - type of data

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_cbsendh( buf, dest, type, len, datatype )
void *buf;
int  dest, type, len, datatype;
{
SENDSYNCNOMEM(type,buf,len,dest,datatype);
return len;
}

/*@
    MPI_cnsendh - Nonblocking contiguous heterogeneous send

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   len  - length in bytes of message
.   datatype - type of data

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_cnsendh( buf, dest, type, len, datatype )
void *buf;
int  dest, type, len, datatype;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_SEND,_MPI_CONTIG);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_SEND
p->buf     = buf;
p->type    = type;
p->len     = len;
p->partner = dest;
#endif
SENDASYNCNOMEM(type,buf,len,dest,datatype,p->sid);
return (int)p;
}

/*@
    MPI_cbrecvh - Blocking contiguous heterogeneous receive

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message
.   datatype - type of data

    Returns:
    Number of bytes actually received.
@*/
int MPI_cbrecvh( buf, source, type, maxlen, datatype )
void *buf;
int  source, type, maxlen, datatype;
{
RECVSYNCNOMEM(type,buf,maxlen,datatype);
return RECVLEN();
}

/*@
    MPI_cnrecvh - Nonblocking contiguous heterogeneous receive

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message
.   datatype - type of data

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_cnrecvh( buf, source, type, maxlen, datatype )
void *buf;
int  source, type, maxlen, datatype;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_RECV,_MPI_CONTIG);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_RECV
p->buf     = buf;
p->type    = type;
p->len     = maxlen;
p->partner = source;
#endif
RECVASYNCNOMEM(type,buf,maxlen,datatype,p->rid);
return (int)p;
}

/*@
    MPI_cbsendhrr - Blocking contiguous heterogeneous send for ready receivers

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   len  - length in bytes of message
.   datatype - type of data

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_cbsendhrr( buf, dest, type, len, datatype )
void *buf;
int  dest, type, len, datatype;
{
SENDSYNCNOMEMFORCE(type,buf,len,dest,datatype);
return len;
}

/*@
    MPI_cnsendhrr - Nonblocking contiguous heterogeneous 
                    send for ready receivers

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   len  - length in bytes of message
.   datatype - type of data

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_cnsendhrr( buf, dest, type, len, datatype )
void *buf;
int  dest, type, len, datatype;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_SEND,_MPI_CONTIG);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_SEND
p->buf     = buf;
p->type    = type;
p->len     = len;
p->partner = dest;
#endif
SENDASYNCNOMEMFORCE(type,buf,len,dest,datatype,p->sid);
return (int)p;
}

/*@
    MPI_cbrecvhrr - Blocking contiguous heterogeneous receive 
                    for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message
.   datatype - type of data

    Returns:
    Number of bytes actually received.
@*/
int MPI_cbrecvhrr( buf, source, type, maxlen, datatype )
void *buf;
int  source, type, maxlen, datatype;
{
RECVSYNCNOMEMFORCE(type,buf,maxlen,datatype);
return RECVLEN();
}

/*@
    MPI_cnrecvhrr - Nonblocking contiguous heterogeneous 
                    receive for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message
.   datatype - type of data

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_cnrecvhrr( buf, source, type, maxlen, datatype )
void *buf;
int  source, type, maxlen, datatype;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_RECV,_MPI_CONTIG);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_RECV
p->buf     = buf;
p->type    = type;
p->len     = maxlen;
p->partner = source;
#endif
RECVASYNCNOMEMFORCE(type,buf,maxlen,datatype,p->rid);
return (int)p;
}

/*@
    MPI_csendh - Friendly contiguous heterogeneous send

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to send
.   dest - destination processor
.   type - message type
.   len  - length in bytes of message
.   datatype - type of data

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_csendh( mode, buf, dest, type, len, datatype )
char *mode;
void *buf;
int  dest, type, len, datatype;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_cbsendh(buf,dest,type,len,datatype);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_cnsendh(buf,dest,type,len,datatype);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_cbsendh(buf,dest,type,len,datatype);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_crecvh - Friendly contiguous heterogeneous receive

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message
.   datatype - type of data

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_crecvh( mode, buf, source, type, maxlen,datatype )
char *mode;
void *buf;
int  source, type, maxlen,datatype;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_cbrecvh(buf,source,type,maxlen,datatype);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_cnrecvh(buf,source,type,maxlen,datatype);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_cbrecvh(buf,source,type,maxlen,datatype);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_csendhrr - Friendly contiguous heterogeneous send for ready 
                   receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to send
.   dest - destination processor
.   type - message type
.   len  - length in bytes of message
.   datatype - type of data

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_csendhrr( mode, buf, dest, type, len, datatype )
char *mode;
void *buf;
int  dest, type, len, datatype;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_cbsendhrr(buf,dest,type,len,datatype);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_cnsendhrr(buf,dest,type,len,datatype);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_cbsendhrr(buf,dest,type,len,datatype);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_crecvhrr - Friendly contiguous heterogeneous 
                   receive for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   maxlen  - maximum length in bytes of message
.   datatype - type of data

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_crecvhrr( mode, buf, source, type, maxlen, datatype )
char *mode;
void *buf;
int  source, type, maxlen, datatype;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_cbrecvhrr(buf,source,type,maxlen,datatype);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_cnrecvhrr(buf,source,type,maxlen,datatype);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_cbrecvhrr(buf,source,type,maxlen,datatype);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}
