#include "../include/mpi.h"

#define _MPI_glen(desc,bcnt) 1
extern _MPI_pendings *_MPI_GetP( );
extern void _MPI_FreeP( );

/*@
    MPI_gbsend - Blocking send with arbitrary gather

    Input Parameters:
.   dest - destination
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_gbsend( dest, type, desc, bcnt )
int         dest, type, bcnt;
MPI_DATAVEC *desc;
{
    MSGBVEC bv;
    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    SENDVSYNC(type,&bv,dest);
    return _MPI_glen(desc,bcnt);
    }

/*@
    MPI_gnsend - Nonblocking send with arbitrary gather

    Input Parameters:
.   dest - destination
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_gnsend( dest, type, desc, bcnt )
int         dest, type, bcnt;
MPI_DATAVEC *desc;
{
    _MPI_pendings *p;
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    p  = _MPI_GetP(_MPI_SEND,_MPI_GENERAL);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_SEND
    p->buf     = 0;
    p->type    = type;
    p->len     = _MPI_glen(desc,bcnt);
    p->partner = dest;
    p->bv      = bv;
#endif

    SENDVASYNC(type,&bv,dest,p->sid);
    return (int)p;
}

/*@
    MPI_gbrecv - Blocking receive into buffer with arbitrary scatter

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Number of bytes actually received.
@*/
int MPI_gbrecv( source, type, desc, bcnt )
int         source, type, bcnt;
MPI_DATAVEC *desc;
{
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    RECVVSYNC(type,&bv);
    return RECVLEN();
}

/*@
    MPI_gnrecv - Nonblocking receive into buffer with arbitrary scatter

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_gnrecv( source, type, desc, bcnt )
int         source, type, bcnt;
MPI_DATAVEC *desc;
{
    _MPI_pendings *p;
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    p  = _MPI_GetP(_MPI_RECV,_MPI_GENERAL);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_RECV
    p->type    = type;
    p->len     = _MPI_glen(desc,bcnt);
    p->partner = source;
    p->bv      = bv;
#endif
    RECVVASYNC(type,&bv,p->rid);
    return (int)p;
}

/*@
    MPI_gbsendrr - Blocking send with arbitrary gather for ready receivers

    Input Parameters:
.   dest - destination
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_gbsendrr( dest, type, desc, bcnt )
int         dest, type, bcnt;
MPI_DATAVEC *desc;
{
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    SENDVSYNCFORCE(type,&bv,dest);
    return _MPI_glen(desc,bcnt);
}

/*@
    MPI_gnsendrr - Nonblocking send with arbitrary gather for ready receivers

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_gnsendrr( dest, type, desc, bcnt )
int         dest, type, bcnt;
MPI_DATAVEC *desc;
{
    _MPI_pendings *p;
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    p  = _MPI_GetP(_MPI_SEND,_MPI_GENERAL);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_SEND
    p->type    = type;
    p->len     = _MPI_glen(desc,bcnt);
    p->partner = dest;
    p->bv      = bv;
#endif
    SENDVASYNCFORCE(type,&bv,dest,p->sid);
    return (int)p;
}

/*@
    MPI_gbrecvrr - Blocking receive into buffer with arbitrary scatter 
                   for ready receivers

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Number of bytes actually received.
@*/
int MPI_gbrecvrr( source, type, desc, bcnt )
int         source, type, bcnt;
MPI_DATAVEC *desc;
{
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    RECVVSYNCFORCE(type,&bv);
    return RECVLEN();
}

/*@
    MPI_gnrecvrr - Nonblocking receive into buffer with arbitrary scatter 
    for ready receivers

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_gnrecvrr( source, type, desc, bcnt )
int         source, type, bcnt;
MPI_DATAVEC *desc;
{
    _MPI_pendings *p;
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    p  = _MPI_GetP(_MPI_RECV,_MPI_GENERAL);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_RECV
    p->type    = type;
    p->len     = _MPI_glen(desc,bcnt);
    p->partner = source;
    p->bv      = bv;
#endif
    RECVVASYNCFORCE(type,&bv,p->rid);
    return (int)p;
}

/*@
    MPI_gsend - Draft standard send with arbitrary gather

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   dest - destination processor
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_gsend( mode, dest, type, desc, bcnt )
char        *mode;
int         dest, type, bcnt;
MPI_DATAVEC *desc;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_gbsend(dest,type,desc,bcnt);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_gnsend(dest,type,desc,bcnt);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_gbsend(dest,type,desc,bcnt);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_grecv - Draft standard receive into buffer with arbitrary scatter

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_grecv( mode, source, type, desc, bcnt )
char        *mode;
int         source, type, bcnt;
MPI_DATAVEC *desc;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_gbrecv(source,type,desc,bcnt);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_gnrecv(source,type,desc,bcnt);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_gbrecv(source,type,desc,bcnt);
	source = MPI_infos();
	if (MPI_gbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_gsendrr - Friendly send with arbitrary gather for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   dest - destination processor
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_gsendrr( mode, dest, type, desc, bcnt )
char        *mode;
int         dest, type, bcnt;
MPI_DATAVEC *desc;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_gbsendrr(dest,type,desc,bcnt);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_gnsendrr(dest,type,desc,bcnt);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_gbsendrr(dest,type,desc,bcnt);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_grecvrr - Friendly receive into buffer with arbitrary scatter
                  for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_grecvrr( mode, source, type, desc, bcnt )
char        *mode;
int         source, type, bcnt;
MPI_DATAVEC *desc;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_gbrecvrr(source,type,desc,bcnt);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_gnrecvrr(source,type,desc,bcnt);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_gbrecvrr(source,type,desc,bcnt);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_gbsendh - Blocking send with arbitrary gather

    Input Parameters:
.   dest - destination
.   type - message type
.   lenblk  - size in bytes of each data block
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_gbsendh( dest, type, desc, bcnt )
int          dest, type, bcnt;
MPI_HDATAVEC *desc;
{
    MSGBVEC bv;
    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    SENDVSYNC(type,&bv,dest);
    return _MPI_glen(desc,bcnt);
}

/*@
    MPI_gnsendh - Nonblocking heterogeneous send with arbitrary gather

    Input Parameters:
.   dest - destination
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_gnsendh( dest, type, desc, bcnt )
int          dest, type, bcnt;
MPI_HDATAVEC *desc;
{
    _MPI_pendings *p;
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    p  = _MPI_GetP(_MPI_SEND,_MPI_GENERAL);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_SEND
    p->type    = type;
    p->len     = _MPI_glen(desc,bcnt);
    p->partner = dest;
    p->bv      = bv;
#endif
    SENDVASYNC(type,&bv,dest,p->sid);
    return (int)p;
}

/*@
    MPI_gbrecvh - Blocking heterogeneous receive into buffer with arbitrary
                  scatter

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_HDATAVEC format
.   bcnt - number of blocks

    Returns:
    Number of bytes actually received.
@*/
int MPI_gbrecvh( source, type, desc, bcnt )
int          source, type, bcnt;
MPI_HDATAVEC *desc;
{
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    RECVVSYNC(type,&bv);
    return RECVLEN();
}

/*@
    MPI_gnrecvh - Nonblocking heterogeneous receive into buffer with arbitrary
                  scatter

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_gnrecvh( source, type, desc, bcnt )
int          source, type, bcnt;
MPI_HDATAVEC *desc;
{
    _MPI_pendings *p;
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    p  = _MPI_GetP(_MPI_RECV,_MPI_GENERAL);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_RECV
    p->type    = type;
    p->len     = _MPI_glen(desc,bcnt);
    p->partner = source;
    p->bv      = bv;
#endif
    RECVVASYNC(type,&bv,p->rid);
    return (int)p;
}

/*@
    MPI_gbsendhrr - Blocking heterogeneous send with arbitrary gather 
                    for ready receivers

    Input Parameters:
.   dest - destination
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_gbsendhrr( dest, type, desc, bcnt )
int          dest, type, bcnt;
MPI_HDATAVEC *desc;
{
    MSGBVEC bv;
    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    SENDVSYNCFORCE(type,&bv,dest);
    return _MPI_glen(lenblk,nblks);
}

/*@
    MPI_gnsendhrr - Nonblocking heterogeneous 
                    send with arbitrary gather for ready receivers

    Input Parameters:
.   dest - destination
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_gnsendhrr( dest, type, desc, bcnt )
int          dest, type, bcnt;
MPI_HDATAVEC *desc;
{
    _MPI_pendings *p;
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    p  = _MPI_GetP(_MPI_SEND,_MPI_GENERAL);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_SEND
    p->type    = type;
    p->len     = _MPI_glen(desc,bcnt);
    p->partner = dest;
    p->bv      = bv;
#endif
    SENDVASYNCFORCE(type,&bv,dest,p->sid);
    return (int)p;
}

/*@
    MPI_gbrecvhrr - Blocking heterogeneous receive into buffer with arbitrary
                    scatter for ready receivers

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Number of bytes actually received.
@*/
int MPI_gbrecvhrr( source, type, desc, bcnt )
int          source, type, bcnt;
MPI_HDATAVEC *desc;
{
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    RECVVSYNCFORCE(type,&bv);
    return RECVLEN();
}

/*@
    MPI_gnrecvhrr - Nonblocking heterogeneous receive into buffer with 
                    arbitrary scatter for ready receivers

    Input Parameters:
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_gnrecvhrr( source, type, desc, bcnt )
int          source, type, bcnt;
MPI_HDATAVEC *desc;
{
    _MPI_pendings *p;
    MSGBVEC bv;

    bv.maxcnt   = bcnt;
    bv.cnt      = bcnt;
    bv.BV       = (struct iovec *)desc;
    bv.datatype = (int *)0;

    p  = _MPI_GetP(_MPI_RECV,_MPI_GENERAL);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_RECV
    p->type    = type;
    p->len     = _MPI_glen(desc,bcnt);
    p->partner = source;
    p->bv      = bv;
#endif
    RECVVASYNCFORCE(type,&bv,p->rid);
    return (int)p;
}

/*@
    MPI_gsendh - Friendly heterogeneous send with arbitrary gather

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   dest - destination processor
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_gsendh( mode, dest, type, desc, bcnt )
char         *mode;
int          dest, type, bcnt;
MPI_HDATAVEC *desc;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_gbsendh(dest,type,desc,bcnt);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_gnsendh(dest,type,desc,bcnt);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_gbsendh(dest,type,desc,bcnt);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_grecvh - Friendly heterogeneous receive into buffer with 
                 arbitrary scatter

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_grecvh( mode, source, type, desc, bcnt )
char         *mode;
int          source, type, bcnt;
MPI_HDATAVEC *desc;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_gbrecvh(source,type,desc,bcnt);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_gnrecvh(source,type,desc,bcnt);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_gbrecvh(source,type,desc,bcnt);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_gsendhrr - Friendly heterogeneous send with arbitrary scatter
                   for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   dest - destination processor
.   type - message type
.   desc - description of data to be gathered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_gsendhrr( mode, dest, type, desc, bcnt )
char         *mode;
int          dest, type, bcnt;
MPI_HDATAVEC *desc;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_gbsendhrr(dest,type,desc,bcnt);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_gnsendhrr(dest,type,desc,bcnt);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_gbsendhrr(dest,type,desc,bcnt);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_grecvhrr - Friendly heterogeneous receive into a buffer with 
                   arbitrary scatter for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   source - sending processor
.   type - message type
.   desc - description of data to be scattered, in MPI_DATAVEC format
.   bcnt - number of blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_grecvhrr( mode, source, type, desc, bcnt )
char         *mode;
int          source, type, bcnt;
MPI_HDATAVEC *desc;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_gbrecvhrr(source,type,desc,bcnt);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_gnrecvhrr(source,type,desc,bcnt);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_gbrecvhrr(source,type,desc,bcnt);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}
