#include "../include/mpi.h"

#define _MPI_slen(lenblk,nblks) (lenblk*nblks)
extern _MPI_pendings *_MPI_GetP( );
extern void _MPI_FreeP( );

/*@
    MPI_sbsend - Blocking send with constant stride

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_sbsend( buf, dest, type, lenblk,stride,nblks )
void *buf;
int  dest, type, lenblk,stride,nblks;
{
SENDVECSYNC(type,buf,lenblk,stride,nblks,dest,MSG_OTHER);
return _MPI_slen(lenblk,nblks);
}

/*@
    MPI_snsend - Nonblocking send with constant stride

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_snsend( buf, dest, type, lenblk,stride,nblks )
void *buf;
int  dest, type, lenblk,stride,nblks;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_SEND,_MPI_STRIDE);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_SEND
p->buf     = buf;
p->type    = type;
p->len     = _MPI_slen(lenblk,nblks);
p->lenblk  = lenblk;
p->stride  = stride;
p->nblks   = nblks;
p->partner = dest;
#endif
SENDVECASYNC(type,buf,lenblk,stride,nblks,dest,MSG_OTHER,p->sid);
return (int)p;
}

/*@
    MPI_sbrecv - Blocking receive into buffer with constant stride

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks

    Returns:
    Number of bytes actually received.
@*/
int MPI_sbrecv( buf, source, type, lenblk,stride,nblks )
void *buf;
int  source, type, lenblk,stride,nblks;
{
RECVVECSYNC(type,buf,lenblk,stride,nblks,MSG_OTHER);
return RECVLEN();
}

/*@
    MPI_snrecv - Nonblocking receive into buffer with constant stride

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_snrecv( buf, source, type, lenblk,stride,nblks )
void *buf;
int  source, type, lenblk,stride,nblks;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_RECV,_MPI_STRIDE);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_RECV
p->buf     = buf;
p->type    = type;
p->len     = _MPI_slen(lenblk,nblks);
p->lenblk  = lenblk;
p->stride  = stride;
p->nblks   = nblks;
p->partner = source;
#endif
RECVVECASYNC(type,buf,lenblk,stride,nblks,MSG_OTHER,p->rid);
return (int)p;
}

/*@
    MPI_sbsendrr - Blocking send with constant stride for ready receivers

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_sbsendrr( buf, dest, type, lenblk,stride,nblks )
void *buf;
int  dest, type, lenblk,stride,nblks;
{
SENDVECSYNCFORCE(type,buf,lenblk,stride,nblks,dest,MSG_OTHER);
return _MPI_slen(lenblk,nblks);
}

/*@
    MPI_snsendrr - Nonblocking send with constant stride for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_snsendrr( buf, dest, type, lenblk,stride,nblks )
void *buf;
int  dest, type, lenblk,stride,nblks;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_SEND,_MPI_STRIDE);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_SEND
p->buf     = buf;
p->type    = type;
p->len     = _MPI_slen(lenblk,nblks);
p->lenblk  = lenblk;
p->stride  = stride;
p->nblks   = nblks;
p->partner = dest;
#endif
SENDVECASYNCFORCE(type,buf,lenblk,stride,nblks,dest,MSG_OTHER,p->sid);
return (int)p;
}

/*@
    MPI_sbrecvrr - Blocking receive into buffer with constant stride for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks

    Returns:
    Number of bytes actually received.
@*/
int MPI_sbrecvrr( buf, source, type, lenblk,stride,nblks )
void *buf;
int  source, type, lenblk,stride,nblks;
{
RECVVECSYNCFORCE(type,buf,lenblk,stride,nblks,MSG_OTHER);
return RECVLEN();
}

/*@
    MPI_snrecvrr - Nonblocking receive into buffer with constant stride for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_snrecvrr( buf, source, type, lenblk,stride,nblks )
void *buf;
int  source, type, lenblk,stride,nblks;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_RECV,_MPI_STRIDE);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_RECV
p->buf     = buf;
p->type    = type;
p->len     = _MPI_slen(lenblk,nblks);
p->lenblk  = lenblk;
p->stride  = stride;
p->nblks   = nblks;
p->partner = source;
#endif
RECVVECASYNCFORCE(type,buf,lenblk,stride,nblks,MSG_OTHER,p->rid);
return (int)p;
}

/*@
    MPI_ssend - Draft standard send with constant stride

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to send
.   dest - destination processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_ssend( mode, buf, dest, type, lenblk,stride,nblks )
char *mode;
void *buf;
int  dest, type, lenblk,stride,nblks;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_sbsend(buf,dest,type,lenblk,stride,nblks);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_snsend(buf,dest,type,lenblk,stride,nblks);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_sbsend(buf,dest,type,lenblk,stride,nblks);
	if (MPI_sbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_srecv - Draft standard receive into buffer with constant stride

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_srecv( mode, buf, source, type, lenblk,stride,nblks )
char *mode;
void *buf;
int  source, type, lenblk,stride,nblks;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_sbrecv(buf,source,type,lenblk,stride,nblks);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_snrecv(buf,source,type,lenblk,stride,nblks);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_sbrecv(buf,source,type,lenblk,stride,nblks);
	source = MPI_infos();
	if (MPI_sbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_ssendrr - Friendly send with constant stride for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to send
.   dest - destination processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_ssendrr( mode, buf, dest, type, lenblk,stride,nblks )
char *mode;
void *buf;
int  dest, type, lenblk,stride,nblks;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_sbsendrr(buf,dest,type,lenblk,stride,nblks);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_snsendrr(buf,dest,type,lenblk,stride,nblks);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_sbsendrr(buf,dest,type,lenblk,stride,nblks);
	if (MPI_sbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_srecvrr - Friendly receive into buffer with constant stride
                  for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_srecvrr( mode, buf, source, type, lenblk,stride,nblks )
char *mode;
void *buf;
int  source, type, lenblk,stride,nblks;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_sbrecvrr(buf,source,type,lenblk,stride,nblks);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_snrecvrr(buf,source,type,lenblk,stride,nblks);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_sbrecvrr(buf,source,type,lenblk,stride,nblks);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_sbsendh - Blocking heterogeneous send with constant stride

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks
.   datatype - type of data

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_sbsendh( buf, dest, type, lenblk,stride,nblks, datatype )
void *buf;
int  dest, type, lenblk,stride,nblks, datatype;
{
    SENDVECSYNC(type,buf,lenblk,stride,nblks,dest,datatype);
    return _MPI_slen(lenblk,nblks);
}

/*@
    MPI_snsendh - Nonblocking heterogeneous send with constant stride

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks
.   datatype - type of data

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_snsendh( buf, dest, type, lenblk,stride,nblks, datatype )
void *buf;
int  dest, type, lenblk,stride,nblks, datatype;
{
    _MPI_pendings *p;
    p  = _MPI_GetP(_MPI_SEND,_MPI_STRIDE);   
    if (!p) {
	MPI_errno = 10;
	return -1;
	}
#ifdef NO_ASYNC_SEND
    p->buf     = buf;
    p->type    = type;
    p->len     = _MPI_slen(lenblk,nblks);
    p->lenblk  = lenblk;
    p->stride  = stride;
    p->nblks   = nblks;
    p->partner = dest;
#endif
    SENDVECASYNC(type,buf,lenblk,stride,nblks,dest,datatype,p->sid);
    return (int)p;
}

/*@
    MPI_sbrecvh - Blocking heterogeneous receive into buffer with constant
                  stride

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks
.   datatype - type of data

    Returns:
    Number of bytes actually received.
@*/
int MPI_sbrecvh( buf, source, type, lenblk,stride,nblks, datatype )
void *buf;
int  source, type, lenblk,stride,nblks, datatype;
{
RECVVECSYNC(type,buf,lenblk,stride,nblks,datatype);
return RECVLEN();
}

/*@
    MPI_snrecvh - Nonblocking heterogeneous receive into buffer with constant
    stride

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks
.   datatype - type of data

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_snrecvh( buf, source, type, lenblk,stride,nblks, datatype )
void *buf;
int  source, type, lenblk,stride,nblks, datatype;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_RECV,_MPI_STRIDE);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_RECV
p->buf     = buf;
p->type    = type;
p->len     = _MPI_slen(lenblk,nblks);
p->lenblk  = lenblk;
p->stride  = stride;
p->nblks   = nblks;
p->partner = source;
#endif
RECVVECASYNC(type,buf,lenblk,stride,nblks,datatype,p->rid);
return (int)p;
}

/*@
    MPI_sbsendhrr - Blocking heterogeneous send with constant stride 
                    for ready receivers

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks
.   datatype - type of data

    Returns:
    Number of bytes sent, or -1 for an error.
@*/
int MPI_sbsendhrr( buf, dest, type, lenblk,stride,nblks, datatype )
void *buf;
int  dest, type, lenblk,stride,nblks, datatype;
{
    SENDVECSYNCFORCE(type,buf,lenblk,stride,nblks,dest,datatype);
    return _MPI_slen(lenblk,nblks);
}

/*@
    MPI_snsendhrr - Nonblocking heterogeneous 
                    send with constant stride for ready receivers

    Input Parameters:
.   buf - buffer to send
.   dest - destination
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks
.   datatype - type of data

    Returns:
    Integer id of send to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_snsendhrr( buf, dest, type, lenblk,stride,nblks, datatype )
void *buf;
int  dest, type, lenblk,stride,nblks, datatype;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_SEND,_MPI_STRIDE);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_SEND
p->buf     = buf;
p->type    = type;
p->len     = _MPI_slen(lenblk,nblks);
p->lenblk  = lenblk;
p->stride  = stride;
p->nblks   = nblks;
p->partner = dest;
#endif
SENDVECASYNCFORCE(type,buf,lenblk,stride,nblks,dest,datatype,p->sid);
return (int)p;
}

/*@
    MPI_sbrecvhrr - Blocking heterogeneous receive into buffer with constant
                    stride for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks
.   datatype - type of data

    Returns:
    Number of bytes actually received.
@*/
int MPI_sbrecvhrr( buf, source, type, lenblk,stride,nblks, datatype )
void *buf;
int  source, type, lenblk,stride,nblks, datatype;
{
RECVVECSYNCFORCE(type,buf,lenblk,stride,nblks,datatype);
return RECVLEN();
}

/*@
    MPI_snrecvhrr - Nonblocking heterogeneous receive into buffer with 
                    constant stride for ready receivers

    Input Parameters:
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks
.   datatype - type of data

    Returns:
    Integer id of receive to be used in MPI_wait, etc., or -1 on error.
@*/
int MPI_snrecvhrr( buf, source, type, lenblk,stride,nblks, datatype )
void *buf;
int  source, type, lenblk,stride,nblks, datatype;
{
_MPI_pendings *p;
p  = _MPI_GetP(_MPI_RECV,_MPI_STRIDE);   
if (!p) {
    MPI_errno = 10;
    return -1;
    }
#ifdef NO_ASYNC_RECV
p->buf     = buf;
p->type    = type;
p->len     = _MPI_slen(lenblk,nblks);
p->lenblk  = lenblk;
p->stride  = stride;
p->nblks   = nblks;
p->partner = source;
#endif
RECVVECASYNCFORCE(type,buf,lenblk,stride,nblks,datatype,p->rid);
return (int)p;
}

/*@
    MPI_ssendh - Friendly heterogeneous send with constant stride

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to send
.   dest - destination processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks
.   datatype - type of data

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_ssendh( mode, buf, dest, type, lenblk,stride,nblks, datatype )
char *mode;
void *buf;
int  dest, type, lenblk,stride,nblks, datatype;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_sbsendh(buf,dest,type,lenblk,stride,nblks,datatype);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_snsendh(buf,dest,type,lenblk,stride,nblks,datatype);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_sbsendh(buf,dest,type,lenblk,stride,nblks,datatype);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_srecvh - Friendly heterogeneous receive into buffer with 
                 constant stride

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks
.   datatype - type of data

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_srecvh( mode, buf, source, type, lenblk,stride,nblks,datatype )
char *mode;
void *buf;
int  source, type, lenblk,stride,nblks,datatype;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_sbrecvh(buf,source,type,lenblk,stride,nblks,datatype);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_snrecvh(buf,source,type,lenblk,stride,nblks,datatype);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_sbrecvh(buf,source,type,lenblk,stride,nblks,datatype);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_ssendhrr - Friendly heterogeneous send with constant stride 
                   for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to send
.   dest - destination processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - number of data blocks
.   datatype - type of data

    Returns:
    If the mode is "nonblocking," returns the integer id of send to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_ssendhrr( mode, buf, dest, type, lenblk,stride,nblks, datatype )
char *mode;
void *buf;
int  dest, type, lenblk,stride,nblks, datatype;
{
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_sbsendhrr(buf,dest,type,lenblk,stride,nblks,datatype);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_snsendhrr(buf,dest,type,lenblk,stride,nblks,datatype);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen = MPI_sbsendhrr(buf,dest,type,lenblk,stride,nblks,datatype);
	if (MPI_cbrecv(&lbuf,0,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}

/*@
    MPI_srecvhrr - Friendly heterogeneous receive into a buffer with 
                   constant stride for ready receivers

    Input Parameters:
.   mode - one of "blocking," "nonblocking," or "synchronous"
.   buf - buffer to receive into
.   source - sending processor
.   type - message type
.   lenblk  - size in bytes of each data block
.   stride  - number of bytes between the start of each data block
.   nblks   - maximum number of data blocks
.   datatype - type of data

    Returns:
    If the mode is "nonblocking," returns the integer id of receive to be 
    used in MPI_wait, etc., or -1 on error.

    Otherwise, returns the actual length of the message in bytes, or -1 on 
    error.
@*/
int MPI_srecvhrr( mode, buf, source, type, lenblk,stride,nblks, datatype )
char *mode;
void *buf;
int  source, type, lenblk,stride,nblks, datatype;
{
    /* Comments:
       We've modified the draft to treat truncation as an error.
       
       Some reasons are:
       If the message is truncated, it isn't clear how the user
       determines that.  For example, should the return length
       be > maxlen if the message was truncated?

       Current vendor implementations (including intel/nx) treat 
       truncation as an error.
     */
    if (strcmp( mode, "blocking" ) == 0) 
	return MPI_sbrecvhrr(buf,source,type,lenblk,stride,nblks,datatype);
    if (strcmp( mode, "nonblocking" ) == 0) 
	return MPI_snrecvhrr(buf,source,type,lenblk,stride,nblks,datatype);
    if (strcmp( mode, "synchronous" ) == 0) {
	int    nlen, lbuf;
	nlen   = MPI_sbrecvhrr(buf,source,type,lenblk,stride,nblks,datatype);
	source = MPI_infos();
	if (MPI_cbsend(&lbuf,source,MPI_AckType,sizeof(int)) < 0)
	    nlen = -1;
	return nlen;
	}
    MPI_errno = 14;
    return -1;
}
